# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from twisted.internet import defer, protocol
from twisted.test import proto_helpers


class FakeFactory(protocol.ClientFactory):
    protocol = proto_helpers.AccumulatingProtocol

    def __init__(self, returnNoProtocol=False):
        self.returnNoProtocol = returnNoProtocol
        self.protocolConnectionMade = defer.Deferred()

    def buildProtocol(self, addr):
        if self.returnNoProtocol:
            return None
        self.proto = protocol.ClientFactory.buildProtocol(self, addr)
        return self.proto


class FakeEndpoint(object):
    def __init__(self, failure=None):
        self.failure = failure
        self.deferred = None

    def connect(self, fac):
        self.factory = fac
        if self.deferred:
            return self.deferred
        if self.failure:
            return defer.fail(self.failure)
        self.proto = fac.buildProtocol(None)
        transport = proto_helpers.StringTransport()
        self.aborted = []
        transport.abortConnection = lambda: self.aborted.append(True)
        self.tlsStarts = []
        transport.startTLS = lambda ctx: self.tlsStarts.append(ctx)
        self.proto.makeConnection(transport)
        self.transport = transport
        return defer.succeed(self.proto)



from twisted.internet import reactor, defer
from twisted.internet.endpoints import clientFromString
from twisted.internet.protocol import Factory
from twisted.protocols.basic import LineReceiver

from txi2p.bob.endpoints import BOBI2PServerEndpoint


class Answer(LineReceiver):
    answers = {
        'How are you?': 'Fine',
        'W007!': 'INORITE?',
        None : "I don't know what you mean",
        }

    def lineReceived(self, line):
        if self.answers.has_key(line):
            self.sendLine(self.answers[line])
        else:
            self.sendLine(self.answers[None])


class AnswerFactory(Factory):
    protocol = Answer


def printDest(port):
    # Print out the I2P Destination to copy to the client
    print 'This server is listening on:'
    print port.getHost().destination
    # Handle Ctl+C
    def shutdown():
        print 'Shutting down'
        port.stopListening()
        d = defer.Deferred()
        reactor.callLater(3, d.callback, 1)
        return d
    reactor.addSystemEventTrigger('before', 'shutdown', shutdown)


bobEndpoint = clientFromString(reactor, 'tcp:127.0.0.1:2827')
endpoint = BOBI2PServerEndpoint(reactor, bobEndpoint, 'keypair.answerserver')
d = endpoint.listen(AnswerFactory())
d.addCallback(printDest)

reactor.run()



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from twisted.internet import interfaces
from zope.interface import implementer

from txi2p.bob.factory import BOBI2PClientFactory, BOBI2PServerFactory


def validateDestination(dest):
    # TODO: Validate I2P domain, B32 etc.
    pass


@implementer(interfaces.IStreamClientEndpoint)
class BOBI2PClientEndpoint(object):
    """
    I2P client endpoint backed by the BOB API.
    """

    def __init__(self, reactor, bobEndpoint, dest,
                 tunnelNick=None,
                 inhost='localhost',
                 inport=None,
                 options=None):
        validateDestination(dest)
        self._reactor = reactor
        self._bobEndpoint = bobEndpoint
        self._dest = dest
        self._tunnelNick = tunnelNick
        self._inhost = inhost
        self._inport = inport
        self._options = options

    def connect(self, fac):
        """
        Connect over I2P.

        The provided factory will have its ``buildProtocol`` method called once
        an I2P client tunnel has been successfully created.

        If the factory's ``buildProtocol`` returns ``None``, the connection
        will immediately close.
        """

        i2pFac = BOBI2PClientFactory(self._reactor, fac, self._bobEndpoint, self._dest,
                                     self._tunnelNick,
                                     self._inhost,
                                     self._inport,
                                     self._options)
        d = self._bobEndpoint.connect(i2pFac)
        # Once the BOB IProtocol is returned, wait for the
        # real IProtocol to be returned after tunnel creation,
        # and pass it to any further registered callbacks.
        d.addCallback(lambda proto: i2pFac.deferred)
        return d


@implementer(interfaces.IStreamServerEndpoint)
class BOBI2PServerEndpoint(object):
    """
    I2P server endpoint backed by the BOB API.
    """

    def __init__(self, reactor, bobEndpoint, keypairPath,
                 tunnelNick=None,
                 outhost='localhost',
                 outport=None,
                 options=None):
        self._reactor = reactor
        self._bobEndpoint = bobEndpoint
        self._keypairPath = keypairPath
        self._tunnelNick = tunnelNick
        self._outhost = outhost
        self._outport = outport
        self._options = options

    def listen(self, fac):
        """
        Listen over I2P.

        The provided factory will have its ``buildProtocol`` method called once
        an I2P server tunnel has been successfully created.

        If the factory's ``buildProtocol`` returns ``None``, the connection
        will immediately close.
        """

        i2pFac = BOBI2PServerFactory(self._reactor, fac, self._bobEndpoint, self._keypairPath,
                                     self._tunnelNick,
                                     self._outhost,
                                     self._outport,
                                     self._options)
        d = self._bobEndpoint.connect(i2pFac)
        # Once the BOB IProtocol is returned, wait for the
        # IListeningPort to be returned after tunnel creation,
        # and pass it to any further registered callbacks.
        d.addCallback(lambda proto: i2pFac.deferred)
        return d



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

import unittest

from txi2p.address import I2PAddress


class TestI2PAddress(unittest.TestCase):
    def test_noPort(self):
        addr = I2PAddress('spam.i2p')
        self.assertEqual((addr.destination, addr.port), ('spam.i2p', None))

    def test_port(self):
        addr = I2PAddress('spam.i2p', 81)
        self.assertEqual((addr.destination, addr.port), ('spam.i2p', 81))

    def test_portString(self):
        addr = I2PAddress('spam.i2p', '81')
        self.assertEqual((addr.destination, addr.port), ('spam.i2p', 81))

    def test_reprWithNoPort(self):
        addr = I2PAddress('spam.i2p')
        self.assertEqual(repr(addr), 'I2PAddress(spam.i2p)')

    def test_reprWithPort(self):
        addr = I2PAddress('spam.i2p', 81)
        self.assertEqual(repr(addr), 'I2PAddress(spam.i2p, 81)')

    def test_reprWithPortString(self):
        addr = I2PAddress('spam.i2p', '81')
        self.assertEqual(repr(addr), 'I2PAddress(spam.i2p, 81)')



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from parsley import makeProtocol
from twisted.internet.interfaces import IListeningPort
from twisted.internet.protocol import Protocol
from zope.interface import implementer

from txi2p import grammar

DEFAULT_INPORT  = 9000
DEFAULT_OUTPORT = 9001


class BOBSender(object):
    def __init__(self, transport):
        self.transport = transport

    def sendClear(self):
        self.transport.write('clear\n')

    def sendGetdest(self):
        self.transport.write('getdest\n')

    def sendGetkeys(self):
        self.transport.write('getkeys\n')

    def sendGetnick(self, tunnelNick):
        self.transport.write('getnick %s\n' % tunnelNick)

    def sendInhost(self, inhost):
        self.transport.write('inhost %s\n' % inhost)

    def sendInport(self, inport):
        self.transport.write('inport %s\n' % inport)

    def sendList(self):
        self.transport.write('list\n')

    def sendNewkeys(self):
        self.transport.write('newkeys\n')

    def sendOption(self):
        self.transport.write('option\n')

    def sendOuthost(self, outhost):
        self.transport.write('outhost %s\n' % outhost)

    def sendOutport(self, outport):
        self.transport.write('outport %s\n' % outport)

    def sendQuiet(self):
        self.transport.write('quiet\n')

    def sendQuit(self):
        self.transport.write('quit\n')

    def sendSetkeys(self, keys):
        self.transport.write('setkeys %s\n' % keys)

    def sendSetnick(self, tunnelNick):
        self.transport.write('setnick %s\n' % tunnelNick)

    def sendShow(self):
        self.transport.write('show\n')

    def sendShowprops(self):
        self.transport.write('showprops\n')

    def sendStart(self):
        self.transport.write('start\n')

    def sendStatus(self, tunnelNick):
        self.transport.write('status %s\n' % tunnelNick)

    def sendStop(self):
        self.transport.write('stop\n')

    def sendVerify(self, key):
        self.transport.write('verify %s\n' % key)

    def sendVisit(self):
        self.transport.write('visit\n')


class BOBReceiver(object):
    currentRule = 'State_init'

    def __init__(self, sender):
        self.sender = sender
        self.tunnelExists = False

    def prepareParsing(self, parser):
        # Store the factory for later use
        self.factory = parser.factory

    def finishParsing(self, reason):
        if self.currentRule != 'State_quit':
            self.factory.bobConnectionFailed(reason)

    def initBOB(self, version):
        self.sender.sendList()
        self.currentRule = 'State_list'

    def processTunnelList(self, tunnels):
        # Default port offset is at the end of the tunnels list
        offset = 2*(len(tunnels))
        if self.factory.tunnelNick:
            for i in range(0, len(tunnels)):
                if tunnels[i]['nickname'] == self.factory.tunnelNick:
                    self.tunnelExists = True
                    self.tunnelRunning = tunnels[i]['running']
                    offset = 2*i
                    # The tunnel will be removed by the Factory
                    # that created it.
                    self.factory.removeTunnelWhenFinished = False
                    break
        else:
            self.factory.tunnelNick = 'txi2p-%d' % (len(tunnels) + 1)
        # If the in/outport were not user-configured, set them.
        if not (hasattr(self.factory, 'inport') and self.factory.inport):
            self.factory.inport = DEFAULT_INPORT + offset
        if not (hasattr(self.factory, 'outport') and self.factory.outport):
            self.factory.outport = DEFAULT_OUTPORT + offset

    def setnick(self, success, info):
        if success:
            if hasattr(self.factory, 'keypair') and self.factory.keypair: # If a keypair was provided, use it
                self.sender.sendSetkeys(self.factory.keypair)
                self.currentRule = 'State_setkeys'
            else: # Get a new keypair
                self.sender.sendNewkeys()
                self.currentRule = 'State_newkeys'
        else:
            print 'setnick ERROR: %s' % info

    def setkeys(self, success, info):
        if success:
            # Update the local Destination
            self.sender.sendGetdest()
            self.currentRule = 'State_getdest'
        else:
            print 'setkeys ERROR: %s' % info

    def newkeys(self, success, info):
        if success:
            # Save the new local Destination
            self.factory.localDest = info
            # Get the new keypair
            self.sender.sendGetkeys()
            self.currentRule = 'State_getkeys'
        else:
            print 'newkeys ERROR: %s' % info

    def quit(self, success, info):
        pass


class I2PClientTunnelCreatorBOBReceiver(BOBReceiver):
    def list(self, success, info, data):
        if success:
            self.processTunnelList(data)
            if self.tunnelExists:
                self.sender.sendGetnick(self.factory.tunnelNick)
                self.currentRule = 'State_getnick'
            else:
                # Set tunnel nickname (and update keypair/localDest state)
                self.sender.sendSetnick(self.factory.tunnelNick)
                self.currentRule = 'State_setnick'
        else:
            print 'list ERROR: %s' % info

    def getnick(self, success, info):
        if success:
            if self.tunnelRunning:
                self.sender.sendStop()
                self.currentRule = 'State_stop'
            else:
                # Update the local Destination
                self.sender.sendGetdest()
                self.currentRule = 'State_getdest'
        else:
            print 'ERROR: %s' % info

    def stop(self, success, info):
        if success:
            # Update the local Destination
            self.sender.sendGetdest()
            self.currentRule = 'State_getdest'
        else:
            print 'stop ERROR: %s' % info

    def getdest(self, success, info):
        if success:
            # Save the local Destination
            self.factory.localDest = info
            if self.tunnelExists:
                # Get the keypair
                self.sender.sendGetkeys()
                self.currentRule = 'State_getkeys'
            else:
                self._setInhost()
        else:
            print 'getdest ERROR: %s' % info

    def getkeys(self, success, info):
        if success:
            # Save the keypair
            self.factory.keypair = info
            self._setInhost()
        else:
            print 'getkeys ERROR: %s' % info

    def _setInhost(self):
        self.sender.sendInhost(self.factory.inhost)
        self.currentRule = 'State_inhost'

    def inhost(self, success, info):
        if success:
            self.sender.sendInport(self.factory.inport)
            self.currentRule = 'State_inport'
        else:
            if info in ['tunnel is active',
                        'tunnel shutting down']:
                # Try again. TODO: Limit retries
                self.sender.sendInhost(self.factory.inhost)
            else:
                print 'inhost ERROR: %s' % info

    def inport(self, success, info):
        if success:
            self.sender.sendStart()
            self.currentRule = 'State_start'
        else:
            print 'inport ERROR: %s' % info

    def start(self, success, info):
        if success:
            print "Client tunnel started"
            self.factory.i2pTunnelCreated()
            self.sender.sendQuit()
            self.currentRule = 'State_quit'
        else:
            print 'start ERROR: %s' % info


class I2PServerTunnelCreatorBOBReceiver(BOBReceiver):
    def list(self, success, info, data):
        if success:
            self.processTunnelList(data)
            if self.tunnelExists:
                self.sender.sendGetnick(self.factory.tunnelNick)
                self.currentRule = 'State_getnick'
            else:
                # Set tunnel nickname (and update keypair/localDest state)
                self.sender.sendSetnick(self.factory.tunnelNick)
                self.currentRule = 'State_setnick'
        else:
            print 'list ERROR: %s' % info

    def getnick(self, success, info):
        if success:
            if self.tunnelRunning:
                self.sender.sendStop()
                self.currentRule = 'State_stop'
            else:
                # Update the local Destination
                self.sender.sendGetdest()
                self.currentRule = 'State_getdest'
        else:
            print 'getnick ERROR: %s' % info

    def stop(self, success, info):
        if success:
            # Update the local Destination
            self.sender.sendGetdest()
            self.currentRule = 'State_getdest'
        else:
            print 'stop ERROR: %s' % info

    def getdest(self, success, info):
        if success:
            # Save the local Destination
            self.factory.localDest = info
            self._setOuthost()
        else:
            print 'getdest ERROR: %s' % info

    def getkeys(self, success, info):
        if success:
            # Save the keypair
            self.factory.keypair = info
            self._setOuthost()
        else:
            print 'getkeys ERROR: %s' % info

    def _setOuthost(self):
        self.sender.sendOuthost(self.factory.outhost)
        self.currentRule = 'State_outhost'

    def outhost(self, success, info):
        if success:
            self.sender.sendOutport(self.factory.outport)
            self.currentRule = 'State_outport'
        else:
            if info in ['tunnel is active',
                        'tunnel shutting down']:
                # Try again. TODO: Limit retries
                self.sender.sendOuthost(self.factory.outhost)
            else:
                print 'outhost ERROR: %s' % info

    def outport(self, success, info):
        if success:
            self.sender.sendStart()
            self.currentRule = 'State_start'
        else:
            print 'outport ERROR: %s' % info

    def start(self, success, info):
        if success:
            print "Server tunnel started"
            self.factory.i2pTunnelCreated()
            self.sender.sendQuit()
            self.currentRule = 'State_quit'
        else:
            print 'start ERROR: %s' % info

class I2PTunnelRemoverBOBReceiver(BOBReceiver):
    def list(self, success, info, data):
        if success:
            self.processTunnelList(data)
            if self.tunnelExists:
                # Get tunnel for nickname
                self.sender.sendGetnick(self.factory.tunnelNick)
                self.currentRule = 'State_getnick'
            else:
                # Tunnel already removed
                pass
        else:
            print 'list ERROR: %s' % info

    def getnick(self, success, info):
        if success:
            self.sender.sendStop()
            self.currentRule = 'State_stop'
        else:
            print 'getnick ERROR: %s' % info

    def stop(self, success, info):
        if success:
            self.sender.sendClear()
            self.currentRule = 'State_clear'
        else:
            print 'stop ERROR: %s' % info

    def clear(self, success, info):
        if success:
            print 'Tunnel removed'
            self.factory.i2pTunnelRemoved()
            self.sender.sendQuit()
            self.currentRule = 'State_quit'
        else:
            if info in ['tunnel is active',
                        'tunnel shutting down']:
                # Try again. TODO: Limit retries
                self.sender.sendClear()
            else:
                print 'clear ERROR: %s ' % info


# A Protocol for making an I2P client tunnel via BOB
I2PClientTunnelCreatorBOBClient = makeProtocol(
    grammar.bobGrammarSource,
    BOBSender,
    I2PClientTunnelCreatorBOBReceiver)

# A Protocol for making an I2P server tunnel via BOB
I2PServerTunnelCreatorBOBClient = makeProtocol(
    grammar.bobGrammarSource,
    BOBSender,
    I2PServerTunnelCreatorBOBReceiver)

# A Protocol for removing a BOB I2P tunnel
I2PTunnelRemoverBOBClient = makeProtocol(
    grammar.bobGrammarSource,
    BOBSender,
    I2PTunnelRemoverBOBReceiver)


class I2PClientTunnelProtocol(Protocol):
    def __init__(self, wrappedProto, dest):
        self.wrappedProto = wrappedProto
        self.dest = dest

    def connectionMade(self):
        # First line sent must be the Destination to connect to.
        self.transport.write(self.dest + '\n')
        self.wrappedProto.makeConnection(self.transport)

    def dataReceived(self, data):
        # Pass all received data to the wrapped Protocol.
        self.wrappedProto.dataReceived(data)

    def connectionLost(self, reason):
        self.factory.i2pConnectionLost(self.wrappedProto, reason)


class I2PServerTunnelProtocol(Protocol):
    def __init__(self, wrappedProto):
        self.wrappedProto = wrappedProto
        self.peer = None

    def connectionMade(self):
        self.wrappedProto.makeConnection(self.transport)

    def dataReceived(self, data):
        if self.peer:
            # Pass all other data to the wrapped Protocol.
            self.wrappedProto.dataReceived(data)
        else:
            # First line is the peer's Destination.
            # TODO: Return this to the user somehow.
            self.peer = data.split('\n')[0]

    def connectionLost(self, reason):
        self.wrappedProto.connectionLost(reason)


@implementer(IListeningPort)
class I2PListeningPort(object):
    def __init__(self, wrappedPort, factoryWrapper, serverAddr):
        self._wrappedPort = wrappedPort
        self._factoryWrapper = factoryWrapper
        self._serverAddr = serverAddr

    def startListening(self):
        self._wrappedPort.startListening()

    def stopListening(self):
        self._factoryWrapper.stopListening(self._wrappedPort)

    def getHost(self):
        return self._serverAddr



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from twisted.internet.interfaces import IAddress
from twisted.python.util import FancyEqMixin
from zope.interface import implementer


@implementer(IAddress)
class I2PAddress(FancyEqMixin, object):
    """
    An L{I2PAddress} represents the address of an L{I2PEndpoint}.

    @ivar destination: An I2P Destination byte string; for example, b'example.i2p'.
    @type destination: L{bytes}

    @ivar port: (Optional) An integer representing the port number.
    @type port: L{int}
    """
    compareAttributes = ('destination', 'port')

    def __init__(self, destination, port=None):
        self.destination = destination
        self.port = int(port) if port else None


    def __repr__(self):
        if self.port:
            return '%s(%s, %d)' % (
                self.__class__.__name__, self.destination, self.port)
        return '%s(%s)' % (
            self.__class__.__name__, self.destination)


    def __hash__(self):
        return hash((self.destination, self.port))



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

import unittest

from parsley import makeGrammar, ParseError

from txi2p.grammar import bobGrammarSource


bobGrammar = makeGrammar(bobGrammarSource, {})

def stringParserFromRule(grammar, rule):
    def parseString(s):
        return getattr(grammar(s), rule)()
    return parseString


class TestBOBGrammar(unittest.TestCase):
    def _test(self, rule, data, expected):
        parse = stringParserFromRule(bobGrammar, rule)
        result = parse(data)
        self.assertEqual(result, expected)

    def test_BOB_clear(self):
        self._test('BOB_clear', 'OK cleared\n', (True, 'cleared'))
        self._test('BOB_clear', 'ERROR tunnel is active\n', (False, 'tunnel is active'))

    def test_BOB_getdest(self):
        self._test('BOB_getdest', 'OK spam\n', (True, 'spam'))

    def test_BOB_getkeys(self):
        self._test('BOB_getkeys', 'OK spameggs\n', (True, 'spameggs'))

    def test_BOB_list(self):
        spam = {
            'nickname': 'spam',
            'starting': False,
            'running': True,
            'stopping': False,
            'keys': True,
            'quiet': False,
            'inport': '12345',
            'inhost': 'localhost',
            'outport': '23456',
            'outhost': 'localhost'
        }
        eggs = {
            'nickname': 'eggs',
            'starting': False,
            'running': False,
            'stopping': False,
            'keys': True,
            'quiet': False,
            'inport': 'not_set',
            'inhost': 'localhost',
            'outport': 'not_set',
            'outhost': 'localhost'
        }

        self._test('BOB_list', 'OK Listing done\n', (True, 'Listing done', []))
        self._test('BOB_list', 'DATA NICKNAME: spam STARTING: false RUNNING: true STOPPING: false KEYS: true QUIET: false INPORT: 12345 INHOST: localhost OUTPORT: 23456 OUTHOST: localhost\nDATA NICKNAME: eggs STARTING: false RUNNING: false STOPPING: false KEYS: true QUIET: false INPORT: not_set INHOST: localhost OUTPORT: not_set OUTHOST: localhost\nOK Listing done\n', (True, 'Listing done', [spam, eggs]))
        self._test('BOB_list', 'ERROR ni!\n', (False, 'ni!', []))



from sys import stdout
from twisted.internet import reactor
from twisted.internet.endpoints import clientFromString
from twisted.internet.protocol import ClientFactory, Protocol

from txi2p.bob.endpoints import BOBI2PClientEndpoint


class Eepsite(Protocol):
    def connectionMade(self):
        print 'Connection made, sending eepsite request'
        self.transport.write('GET / HTTP/1.1\r\n\r\n')

    def dataReceived(self, data):
        stdout.write(data)

    def connectionLost(self, reason):
        print 'Lost connection.  Reason:', reason
        reactor.stop()


class EepsiteFactory(ClientFactory):
    protocol = Eepsite


bobEndpoint = clientFromString(reactor, 'tcp:127.0.0.1:2827')
endpoint = BOBI2PClientEndpoint(reactor, bobEndpoint, 'stats.i2p')
d = endpoint.connect(EepsiteFactory())

reactor.run()



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from setuptools import setup


with open('README.rst', 'rb') as infile:
    long_description = infile.read()

with open('requirements.txt', 'rb') as infile:
    install_requires = infile.read().split()

setup(
    name='txi2p',
    description='I2P bindings for Twisted',
    long_description=long_description,
    author='str4d',
    author_email='str4d@i2pmail.org',
    url='https://github.com/str4d/txi2p',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Twisted',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 2 :: Only',
        'Topic :: Internet',
    ],
    license='ISC',

    setup_requires=['vcversioner'],
    vcversioner={
        'version_module_paths': ['txi2p/_version.py'],
    },
    install_requires=install_requires,
    packages=['txi2p', 'txi2p.bob', 'txi2p.test', 'txi2p.bob.test'], #+ ['twisted.plugins'],
)

# Make Twisted regenerate the dropin.cache, if possible.  This is necessary
# because in a site-wide install, dropin.cache cannot be rewritten by
# normal users.
try:
    from twisted.plugin import IPlugin, getPlugins
except ImportError:
    pass
else:
    list(getPlugins(IPlugin))



# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from twisted.internet.protocol import ClientFactory
from twisted.test import proto_helpers
from twisted.trial import unittest

from txi2p.bob.protocol import (I2PClientTunnelCreatorBOBClient,
                                I2PServerTunnelCreatorBOBClient,
                                I2PTunnelRemoverBOBClient,
                                I2PClientTunnelProtocol,
                                I2PServerTunnelProtocol,
                                DEFAULT_INPORT, DEFAULT_OUTPORT)


class BOBProtoTestMixin(object):
    def makeProto(self, *a, **kw):
        protoClass = kw.pop('_protoClass', self.protocol)
        fac = ClientFactory(*a, **kw)
        fac.protocol = protoClass
        proto = fac.buildProtocol(None)
        transport = proto_helpers.StringTransport()
        transport.abortConnection = lambda: None
        proto.makeConnection(transport)
        return fac, proto

    def test_initBOBListsTunnels(self):
        fac, proto = self.makeProto()
        proto.dataReceived('BOB 00.00.10\nOK\n')
        self.assertEqual(proto.transport.value(), 'list\n')

    def TODO_test_quitDoesNotErrback(self):
        fac, proto = self.makeProto()
        # Shortcut to end of BOB protocol
        proto.receiver.currentRule = 'State_quit'
        proto._parser._setupInterp()
        proto.dataReceived('OK Bye!\n')
        self.assertTrue(False, 'TODO: Test something') # TODO: Test something

class BOBTunnelCreationMixin(BOBProtoTestMixin):
    def test_defaultNickSetsNick(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = None
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        self.assertEqual(proto.transport.value(), 'setnick txi2p-1\n')

    def test_newNickSetsNick(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        self.assertEqual(proto.transport.value(), 'setnick spam\n')

    def test_nickSetWithKeypair(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.keypair = 'eggs'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), 'setkeys eggs\n')

    def test_destFetchedAfterNickSetWithKeypair(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.keypair = 'eggs'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), 'getdest\n')

    def test_nickSetWithNoKeypair(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), 'newkeys\n')

    def test_keypairFetchedAfterNickSetWithNoKeypair(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The new Destination
        self.assertEqual(proto.transport.value(), 'getkeys\n')

    def test_existingNickGetsNick(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('DATA NICKNAME: spam STARTING: false RUNNING: true STOPPING: false KEYS: true QUIET: false INPORT: 12345 INHOST: localhost OUTPORT: 23456 OUTHOST: localhost\nOK Listing done\n')
        self.assertEqual(proto.transport.value(), 'getnick spam\n')

    def test_stopRequestedForRunningTunnel(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('DATA NICKNAME: spam STARTING: false RUNNING: true STOPPING: false KEYS: true QUIET: false INPORT: 12345 INHOST: localhost OUTPORT: 23456 OUTHOST: localhost\nOK Listing done\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), 'stop\n')

    def test_stopNotRequestedForStoppedTunnel(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('DATA NICKNAME: spam STARTING: false RUNNING: false STOPPING: false KEYS: true QUIET: false INPORT: 12345 INHOST: localhost OUTPORT: 23456 OUTHOST: localhost\nOK Listing done\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertNotEqual(proto.transport.value(), 'stop\n') # TODO: Refactor to test against what is actually expected


class TestI2PClientTunnelCreatorBOBClient(BOBTunnelCreationMixin, unittest.TestCase):
    protocol = I2PClientTunnelCreatorBOBClient

    def test_inhostRequestRepeatedIfActive(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.inhost = 'camelot'
        # Shortcut
        proto.receiver.currentRule = 'State_inhost'
        proto._parser._setupInterp()
        proto.dataReceived('ERROR tunnel is active\n')
        self.assertEqual(proto.transport.value(), 'inhost camelot\n')

    def test_inhostRequestRepeatedIfShuttingDown(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.inhost = 'camelot'
        # Shortcut
        proto.receiver.currentRule = 'State_inhost'
        proto._parser._setupInterp()
        proto.dataReceived('ERROR tunnel shutting down\n')
        self.assertEqual(proto.transport.value(), 'inhost camelot\n')

    def test_inhostSetAfterNickSetWithKeypair(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.keypair = 'eggs'
        fac.inhost = 'camelot'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The Destination
        self.assertEqual(proto.transport.value(), 'inhost camelot\n')

    def test_inhostSetAfterNickSetWithNoKeypair(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.inhost = 'camelot'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The new Destination
        proto.transport.clear()
        proto.dataReceived('OK rubberyeggs\n') # The new keypair
        self.assertEqual(proto.transport.value(), 'inhost camelot\n')

    def test_defaultInportSet(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.inhost = 'camelot'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The new Destination
        proto.transport.clear()
        proto.dataReceived('OK rubberyeggs\n') # The new keypair
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), 'inport %s\n' % DEFAULT_INPORT)

    def test_inportSet(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.inhost = 'camelot'
        fac.inport = '1234'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The new Destination
        proto.transport.clear()
        proto.dataReceived('OK rubberyeggs\n') # The new keypair
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), 'inport 1234\n')

    def test_startRequested(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.inhost = 'camelot'
        fac.inport = '1234'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The new Destination
        proto.transport.clear()
        proto.dataReceived('OK rubberyeggs\n') # The new keypair
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), 'start\n')


class TestI2PServerTunnelCreatorBOBClient(BOBTunnelCreationMixin, unittest.TestCase):
    protocol = I2PServerTunnelCreatorBOBClient

    def test_outhostRequestRepeatedIfActive(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.outhost = 'camelot'
        # Shortcut
        proto.receiver.currentRule = 'State_outhost'
        proto._parser._setupInterp()
        proto.dataReceived('ERROR tunnel is active\n')
        self.assertEqual(proto.transport.value(), 'outhost camelot\n')

    def test_outhostRequestRepeatedIfShuttingDown(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.outhost = 'camelot'
        # Shortcut
        proto.receiver.currentRule = 'State_outhost'
        proto._parser._setupInterp()
        proto.dataReceived('ERROR tunnel shutting down\n')
        self.assertEqual(proto.transport.value(), 'outhost camelot\n')

    def test_outhostSetAfterNickSetWithKeypair(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.keypair = 'eggs'
        fac.outhost = 'camelot'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The Destination
        self.assertEqual(proto.transport.value(), 'outhost camelot\n')

    def test_outhostSetAfterNickSetWithNoKeypair(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.outhost = 'camelot'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The new Destination
        proto.transport.clear()
        proto.dataReceived('OK rubberyeggs\n') # The new keypair
        self.assertEqual(proto.transport.value(), 'outhost camelot\n')

    def test_defaultOutportSet(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.outhost = 'camelot'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The new Destination
        proto.transport.clear()
        proto.dataReceived('OK rubberyeggs\n') # The new keypair
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), 'outport %s\n' % DEFAULT_OUTPORT)

    def test_outportSet(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.outhost = 'camelot'
        fac.outport = '1234'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The new Destination
        proto.transport.clear()
        proto.dataReceived('OK rubberyeggs\n') # The new keypair
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), 'outport 1234\n')

    def test_startRequested(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        fac.outhost = 'camelot'
        fac.outport = '1234'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK shrubbery\n') # The new Destination
        proto.transport.clear()
        proto.dataReceived('OK rubberyeggs\n') # The new keypair
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), 'start\n')


class TestI2PTunnelRemoverBOBClient(BOBProtoTestMixin, unittest.TestCase):
    protocol = I2PTunnelRemoverBOBClient

    def test_noTunnelWithNick(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('OK Listing done\n') # No DATA, no tunnels
        self.assertEqual(proto.transport.value(), '')

    def test_tunnelExistsGetsNick(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('DATA NICKNAME: spam STARTING: false RUNNING: true STOPPING: false KEYS: true QUIET: false INPORT: 12345 INHOST: localhost OUTPORT: 23456 OUTHOST: localhost\nOK Listing done\n')
        self.assertEqual(proto.transport.value(), 'getnick spam\n')

    def test_stopRequested(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('DATA NICKNAME: spam STARTING: false RUNNING: true STOPPING: false KEYS: true QUIET: false INPORT: 12345 INHOST: localhost OUTPORT: 23456 OUTHOST: localhost\nOK Listing done\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), 'stop\n')

    def test_clearRequested(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        proto.dataReceived('BOB 00.00.10\nOK\n')
        proto.transport.clear()
        proto.dataReceived('DATA NICKNAME: spam STARTING: false RUNNING: true STOPPING: false KEYS: true QUIET: false INPORT: 12345 INHOST: localhost OUTPORT: 23456 OUTHOST: localhost\nOK Listing done\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        proto.transport.clear()
        proto.dataReceived('OK HTTP 418\n')
        self.assertEqual(proto.transport.value(), 'clear\n')

    def test_clearRequestRepeatedIfActive(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        # Shortcut
        proto.receiver.currentRule = 'State_clear'
        proto._parser._setupInterp()
        proto.dataReceived('ERROR tunnel is active\n')
        self.assertEqual(proto.transport.value(), 'clear\n')

    def test_clearRequestRepeatedIfShuttingDown(self):
        fac, proto = self.makeProto()
        fac.tunnelNick = 'spam'
        # Shortcut
        proto.receiver.currentRule = 'State_clear'
        proto._parser._setupInterp()
        proto.dataReceived('ERROR tunnel shutting down\n')
        self.assertEqual(proto.transport.value(), 'clear\n')


class TestI2PClientTunnelProtocol(unittest.TestCase):
    def makeProto(self):
        wrappedProto = proto_helpers.AccumulatingProtocol()
        proto = I2PClientTunnelProtocol(wrappedProto, 'spam.i2p')
        transport = proto_helpers.StringTransport()
        transport.abortConnection = lambda: None
        proto.makeConnection(transport)
        return proto

    def test_destRequested(self):
        proto = self.makeProto()
        self.assertEqual(proto.transport.value(), 'spam.i2p\n')

    def test_wrappedProtoConnectionMade(self):
        proto = self.makeProto()
        self.assertEqual(proto.transport, proto.wrappedProto.transport)

    def test_dataPassed(self):
        proto = self.makeProto()
        proto.dataReceived('shrubbery')
        self.assertEqual(proto.wrappedProto.data, 'shrubbery')


class TestI2PServerTunnelProtocol(unittest.TestCase):
    def makeProto(self):
        wrappedProto = proto_helpers.AccumulatingProtocol()
        proto = I2PServerTunnelProtocol(wrappedProto)
        transport = proto_helpers.StringTransport()
        transport.abortConnection = lambda: None
        proto.makeConnection(transport)
        return proto

    def test_wrappedProtoConnectionMade(self):
        proto = self.makeProto()
        self.assertEqual(proto.transport, proto.wrappedProto.transport)

    def test_peerDestStored(self):
        proto = self.makeProto()
        proto.dataReceived('spam.i2p\n')
        self.assertEqual(proto.peer, 'spam.i2p')

    def test_dataAfterPeerDestPassed(self):
        proto = self.makeProto()
        proto.dataReceived('spam.i2p\n')
        proto.dataReceived('shrubbery')
        self.assertEqual(proto.wrappedProto.data, 'shrubbery')
